# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import bytes_with_io
class ResourceFork(KaitaiStruct):
    """The data format of Macintosh resource forks,
    used on Classic Mac OS and Mac OS X/macOS to store additional structured data along with a file's main data (the data fork).
    The kinds of data stored in resource forks include:
    
    * Document resources:
      images, sounds, etc. used by a document
    * Application resources:
      graphics, GUI layouts, localizable strings,
      and even code used by an application, a library, or system files
    * Common metadata:
      custom icons and version metadata that could be displayed by the Finder
    * Application-specific metadata:
      because resource forks follow a common format,
      other applications can store new metadata in them,
      even if the original application does not recognize or understand it
    
    Macintosh file systems (MFS, HFS, HFS+, APFS) support resource forks natively,
    which allows storing resources along with any file.
    Non-Macintosh file systems and protocols have little or no support for resource forks,
    so the resource fork data must be stored in some other way when using such file systems or protocols.
    Various file formats and tools exist for this purpose,
    such as BinHex, MacBinary, AppleSingle, AppleDouble, or QuickTime RezWack.
    In some cases,
    resource forks are stored as plain data in separate files with a .rsrc extension,
    even on Mac systems that natively support resource forks.
    
    On modern Mac OS X/macOS systems,
    resource forks are used far less commonly than on classic Mac OS systems,
    because of compatibility issues with other systems and historical limitations in the format.
    Modern macOS APIs and libraries do not use resource forks,
    and the legacy Carbon API that still used them has been deprecated since OS X 10.8.
    Despite this,
    even current macOS systems still use resource forks for certain purposes,
    such as custom file icons.
    
    .. seealso::
       Inside Macintosh, More Macintosh Toolbox, Resource Manager, Resource Manager Reference, Resource File Format - https://developer.apple.com/library/archive/documentation/mac/pdf/MoreMacintoshToolbox.pdf#page=151
    
    
    .. seealso::
       Inside Macintosh, Volume I, The Resource Manager, Format of a Resource File - http://www.pagetable.com/?p=50
    
    
    .. seealso::
       Source - https://github.com/kreativekorp/ksfl/wiki/Macintosh-Resource-File-Format
    
    
    .. seealso::
       Source - https://github.com/dgelessus/mac_file_format_docs/blob/master/README.md#resource-forks
    """
    SEQ_FIELDS = ["header", "system_data", "application_data"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = ResourceFork.FileHeader(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['system_data']['start'] = self._io.pos()
        self.system_data = self._io.read_bytes(112)
        self._debug['system_data']['end'] = self._io.pos()
        self._debug['application_data']['start'] = self._io.pos()
        self.application_data = self._io.read_bytes(128)
        self._debug['application_data']['end'] = self._io.pos()

    class FileHeader(KaitaiStruct):
        """Resource file header,
        containing the offsets and lengths of the resource data area and resource map.
        """
        SEQ_FIELDS = ["ofs_data_blocks", "ofs_resource_map", "len_data_blocks", "len_resource_map"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs_data_blocks']['start'] = self._io.pos()
            self.ofs_data_blocks = self._io.read_u4be()
            self._debug['ofs_data_blocks']['end'] = self._io.pos()
            self._debug['ofs_resource_map']['start'] = self._io.pos()
            self.ofs_resource_map = self._io.read_u4be()
            self._debug['ofs_resource_map']['end'] = self._io.pos()
            self._debug['len_data_blocks']['start'] = self._io.pos()
            self.len_data_blocks = self._io.read_u4be()
            self._debug['len_data_blocks']['end'] = self._io.pos()
            self._debug['len_resource_map']['start'] = self._io.pos()
            self.len_resource_map = self._io.read_u4be()
            self._debug['len_resource_map']['end'] = self._io.pos()


    class DataBlock(KaitaiStruct):
        """A resource data block,
        as stored in the resource data area.
        
        Each data block stores the data contained in a resource,
        along with its length.
        """
        SEQ_FIELDS = ["len_data", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u4be()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes(self.len_data)
            self._debug['data']['end'] = self._io.pos()


    class ResourceMap(KaitaiStruct):
        """Resource map,
        containing information about the resources in the file and where they are located in the data area.
        """
        SEQ_FIELDS = ["reserved_file_header_copy", "reserved_next_resource_map_handle", "reserved_file_reference_number", "file_attributes", "ofs_type_list", "ofs_names"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['reserved_file_header_copy']['start'] = self._io.pos()
            self.reserved_file_header_copy = ResourceFork.FileHeader(self._io, self, self._root)
            self.reserved_file_header_copy._read()
            self._debug['reserved_file_header_copy']['end'] = self._io.pos()
            self._debug['reserved_next_resource_map_handle']['start'] = self._io.pos()
            self.reserved_next_resource_map_handle = self._io.read_u4be()
            self._debug['reserved_next_resource_map_handle']['end'] = self._io.pos()
            self._debug['reserved_file_reference_number']['start'] = self._io.pos()
            self.reserved_file_reference_number = self._io.read_u2be()
            self._debug['reserved_file_reference_number']['end'] = self._io.pos()
            self._debug['file_attributes']['start'] = self._io.pos()
            self._raw_file_attributes = self._io.read_bytes(2)
            _io__raw_file_attributes = KaitaiStream(BytesIO(self._raw_file_attributes))
            self.file_attributes = ResourceFork.ResourceMap.FileAttributes(_io__raw_file_attributes, self, self._root)
            self.file_attributes._read()
            self._debug['file_attributes']['end'] = self._io.pos()
            self._debug['ofs_type_list']['start'] = self._io.pos()
            self.ofs_type_list = self._io.read_u2be()
            self._debug['ofs_type_list']['end'] = self._io.pos()
            self._debug['ofs_names']['start'] = self._io.pos()
            self.ofs_names = self._io.read_u2be()
            self._debug['ofs_names']['end'] = self._io.pos()

        class FileAttributes(KaitaiStruct):
            """A resource file's attributes,
            as stored in the resource map.
            
            These attributes are sometimes also referred to as resource map attributes,
            because of where they are stored in the file.
            """
            SEQ_FIELDS = ["resources_locked", "reserved0", "printer_driver_multifinder_compatible", "no_write_changes", "needs_compact", "map_needs_write", "reserved1"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['resources_locked']['start'] = self._io.pos()
                self.resources_locked = self._io.read_bits_int_be(1) != 0
                self._debug['resources_locked']['end'] = self._io.pos()
                self._debug['reserved0']['start'] = self._io.pos()
                self.reserved0 = self._io.read_bits_int_be(6)
                self._debug['reserved0']['end'] = self._io.pos()
                self._debug['printer_driver_multifinder_compatible']['start'] = self._io.pos()
                self.printer_driver_multifinder_compatible = self._io.read_bits_int_be(1) != 0
                self._debug['printer_driver_multifinder_compatible']['end'] = self._io.pos()
                self._debug['no_write_changes']['start'] = self._io.pos()
                self.no_write_changes = self._io.read_bits_int_be(1) != 0
                self._debug['no_write_changes']['end'] = self._io.pos()
                self._debug['needs_compact']['start'] = self._io.pos()
                self.needs_compact = self._io.read_bits_int_be(1) != 0
                self._debug['needs_compact']['end'] = self._io.pos()
                self._debug['map_needs_write']['start'] = self._io.pos()
                self.map_needs_write = self._io.read_bits_int_be(1) != 0
                self._debug['map_needs_write']['end'] = self._io.pos()
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bits_int_be(5)
                self._debug['reserved1']['end'] = self._io.pos()

            @property
            def as_int(self):
                """The attributes as a packed integer,
                as they are stored in the file.
                """
                if hasattr(self, '_m_as_int'):
                    return self._m_as_int if hasattr(self, '_m_as_int') else None

                _pos = self._io.pos()
                self._io.seek(0)
                self._debug['_m_as_int']['start'] = self._io.pos()
                self._m_as_int = self._io.read_u2be()
                self._debug['_m_as_int']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_as_int if hasattr(self, '_m_as_int') else None


        class TypeListAndReferenceLists(KaitaiStruct):
            """Resource type list and storage area for resource reference lists in the resource map.
            
            The two parts are combined into a single type here for technical reasons:
            the start of the resource reference list area is not stored explicitly in the file,
            instead it always starts directly after the resource type list.
            The simplest way to implement this is by placing both types into a single `seq`.
            """
            SEQ_FIELDS = ["type_list", "reference_lists"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['type_list']['start'] = self._io.pos()
                self.type_list = ResourceFork.ResourceMap.TypeListAndReferenceLists.TypeList(self._io, self, self._root)
                self.type_list._read()
                self._debug['type_list']['end'] = self._io.pos()
                self._debug['reference_lists']['start'] = self._io.pos()
                self.reference_lists = self._io.read_bytes_full()
                self._debug['reference_lists']['end'] = self._io.pos()

            class TypeList(KaitaiStruct):
                """Resource type list in the resource map."""
                SEQ_FIELDS = ["num_types_m1", "entries"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['num_types_m1']['start'] = self._io.pos()
                    self.num_types_m1 = self._io.read_u2be()
                    self._debug['num_types_m1']['end'] = self._io.pos()
                    self._debug['entries']['start'] = self._io.pos()
                    self.entries = [None] * (self.num_types)
                    for i in range(self.num_types):
                        if not 'arr' in self._debug['entries']:
                            self._debug['entries']['arr'] = []
                        self._debug['entries']['arr'].append({'start': self._io.pos()})
                        _t_entries = ResourceFork.ResourceMap.TypeListAndReferenceLists.TypeList.TypeListEntry(self._io, self, self._root)
                        _t_entries._read()
                        self.entries[i] = _t_entries
                        self._debug['entries']['arr'][i]['end'] = self._io.pos()

                    self._debug['entries']['end'] = self._io.pos()

                class TypeListEntry(KaitaiStruct):
                    """A single entry in the resource type list.
                    
                    Each entry corresponds to exactly one resource reference list.
                    """
                    SEQ_FIELDS = ["type", "num_references_m1", "ofs_reference_list"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['type']['start'] = self._io.pos()
                        self.type = self._io.read_bytes(4)
                        self._debug['type']['end'] = self._io.pos()
                        self._debug['num_references_m1']['start'] = self._io.pos()
                        self.num_references_m1 = self._io.read_u2be()
                        self._debug['num_references_m1']['end'] = self._io.pos()
                        self._debug['ofs_reference_list']['start'] = self._io.pos()
                        self.ofs_reference_list = self._io.read_u2be()
                        self._debug['ofs_reference_list']['end'] = self._io.pos()

                    @property
                    def num_references(self):
                        """The number of resources in the reference list for this type."""
                        if hasattr(self, '_m_num_references'):
                            return self._m_num_references if hasattr(self, '_m_num_references') else None

                        self._m_num_references = ((self.num_references_m1 + 1) % 65536)
                        return self._m_num_references if hasattr(self, '_m_num_references') else None

                    @property
                    def reference_list(self):
                        """The resource reference list for this resource type.
                        """
                        if hasattr(self, '_m_reference_list'):
                            return self._m_reference_list if hasattr(self, '_m_reference_list') else None

                        io = self._parent._parent._io
                        _pos = io.pos()
                        io.seek(self.ofs_reference_list)
                        self._debug['_m_reference_list']['start'] = io.pos()
                        self._m_reference_list = ResourceFork.ResourceMap.TypeListAndReferenceLists.ReferenceList(self.num_references, io, self, self._root)
                        self._m_reference_list._read()
                        self._debug['_m_reference_list']['end'] = io.pos()
                        io.seek(_pos)
                        return self._m_reference_list if hasattr(self, '_m_reference_list') else None


                @property
                def num_types(self):
                    """The number of resource types in this list."""
                    if hasattr(self, '_m_num_types'):
                        return self._m_num_types if hasattr(self, '_m_num_types') else None

                    self._m_num_types = ((self.num_types_m1 + 1) % 65536)
                    return self._m_num_types if hasattr(self, '_m_num_types') else None


            class ReferenceList(KaitaiStruct):
                """A resource reference list,
                as stored in the reference list area.
                
                Each reference list has exactly one matching entry in the resource type list,
                and describes all resources of a single type in the file.
                """
                SEQ_FIELDS = ["references"]
                def __init__(self, num_references, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self.num_references = num_references
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['references']['start'] = self._io.pos()
                    self.references = [None] * (self.num_references)
                    for i in range(self.num_references):
                        if not 'arr' in self._debug['references']:
                            self._debug['references']['arr'] = []
                        self._debug['references']['arr'].append({'start': self._io.pos()})
                        _t_references = ResourceFork.ResourceMap.TypeListAndReferenceLists.ReferenceList.Reference(self._io, self, self._root)
                        _t_references._read()
                        self.references[i] = _t_references
                        self._debug['references']['arr'][i]['end'] = self._io.pos()

                    self._debug['references']['end'] = self._io.pos()

                class Reference(KaitaiStruct):
                    """A single resource reference in a resource reference list."""
                    SEQ_FIELDS = ["id", "ofs_name", "attributes", "ofs_data_block", "reserved_handle"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['id']['start'] = self._io.pos()
                        self.id = self._io.read_s2be()
                        self._debug['id']['end'] = self._io.pos()
                        self._debug['ofs_name']['start'] = self._io.pos()
                        self.ofs_name = self._io.read_u2be()
                        self._debug['ofs_name']['end'] = self._io.pos()
                        self._debug['attributes']['start'] = self._io.pos()
                        self._raw_attributes = self._io.read_bytes(1)
                        _io__raw_attributes = KaitaiStream(BytesIO(self._raw_attributes))
                        self.attributes = ResourceFork.ResourceMap.TypeListAndReferenceLists.ReferenceList.Reference.Attributes(_io__raw_attributes, self, self._root)
                        self.attributes._read()
                        self._debug['attributes']['end'] = self._io.pos()
                        self._debug['ofs_data_block']['start'] = self._io.pos()
                        self.ofs_data_block = self._io.read_bits_int_be(24)
                        self._debug['ofs_data_block']['end'] = self._io.pos()
                        self._io.align_to_byte()
                        self._debug['reserved_handle']['start'] = self._io.pos()
                        self.reserved_handle = self._io.read_u4be()
                        self._debug['reserved_handle']['end'] = self._io.pos()

                    class Attributes(KaitaiStruct):
                        """A resource's attributes,
                        as stored in a resource reference.
                        """
                        SEQ_FIELDS = ["system_reference", "load_into_system_heap", "purgeable", "locked", "protected", "preload", "needs_write", "compressed"]
                        def __init__(self, _io, _parent=None, _root=None):
                            self._io = _io
                            self._parent = _parent
                            self._root = _root if _root else self
                            self._debug = collections.defaultdict(dict)

                        def _read(self):
                            self._debug['system_reference']['start'] = self._io.pos()
                            self.system_reference = self._io.read_bits_int_be(1) != 0
                            self._debug['system_reference']['end'] = self._io.pos()
                            self._debug['load_into_system_heap']['start'] = self._io.pos()
                            self.load_into_system_heap = self._io.read_bits_int_be(1) != 0
                            self._debug['load_into_system_heap']['end'] = self._io.pos()
                            self._debug['purgeable']['start'] = self._io.pos()
                            self.purgeable = self._io.read_bits_int_be(1) != 0
                            self._debug['purgeable']['end'] = self._io.pos()
                            self._debug['locked']['start'] = self._io.pos()
                            self.locked = self._io.read_bits_int_be(1) != 0
                            self._debug['locked']['end'] = self._io.pos()
                            self._debug['protected']['start'] = self._io.pos()
                            self.protected = self._io.read_bits_int_be(1) != 0
                            self._debug['protected']['end'] = self._io.pos()
                            self._debug['preload']['start'] = self._io.pos()
                            self.preload = self._io.read_bits_int_be(1) != 0
                            self._debug['preload']['end'] = self._io.pos()
                            self._debug['needs_write']['start'] = self._io.pos()
                            self.needs_write = self._io.read_bits_int_be(1) != 0
                            self._debug['needs_write']['end'] = self._io.pos()
                            self._debug['compressed']['start'] = self._io.pos()
                            self.compressed = self._io.read_bits_int_be(1) != 0
                            self._debug['compressed']['end'] = self._io.pos()

                        @property
                        def as_int(self):
                            """The attributes as a packed integer,
                            as they are stored in the file.
                            """
                            if hasattr(self, '_m_as_int'):
                                return self._m_as_int if hasattr(self, '_m_as_int') else None

                            _pos = self._io.pos()
                            self._io.seek(0)
                            self._debug['_m_as_int']['start'] = self._io.pos()
                            self._m_as_int = self._io.read_u1()
                            self._debug['_m_as_int']['end'] = self._io.pos()
                            self._io.seek(_pos)
                            return self._m_as_int if hasattr(self, '_m_as_int') else None


                    @property
                    def name(self):
                        """The name (if any) of the resource described by this reference.
                        """
                        if hasattr(self, '_m_name'):
                            return self._m_name if hasattr(self, '_m_name') else None

                        if self.ofs_name != 65535:
                            io = self._root.resource_map.names_with_io._io
                            _pos = io.pos()
                            io.seek(self.ofs_name)
                            self._debug['_m_name']['start'] = io.pos()
                            self._m_name = ResourceFork.ResourceMap.Name(io, self, self._root)
                            self._m_name._read()
                            self._debug['_m_name']['end'] = io.pos()
                            io.seek(_pos)

                        return self._m_name if hasattr(self, '_m_name') else None

                    @property
                    def data_block(self):
                        """The data block containing the data for the resource described by this reference.
                        """
                        if hasattr(self, '_m_data_block'):
                            return self._m_data_block if hasattr(self, '_m_data_block') else None

                        io = self._root.data_blocks_with_io._io
                        _pos = io.pos()
                        io.seek(self.ofs_data_block)
                        self._debug['_m_data_block']['start'] = io.pos()
                        self._m_data_block = ResourceFork.DataBlock(io, self, self._root)
                        self._m_data_block._read()
                        self._debug['_m_data_block']['end'] = io.pos()
                        io.seek(_pos)
                        return self._m_data_block if hasattr(self, '_m_data_block') else None




        class Name(KaitaiStruct):
            """A resource name,
            as stored in the resource name storage area in the resource map.
            
            The resource names are not required to appear in any particular order.
            There may be unused space between and around resource names,
            but in practice they are often contiguous.
            """
            SEQ_FIELDS = ["len_value", "value"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['len_value']['start'] = self._io.pos()
                self.len_value = self._io.read_u1()
                self._debug['len_value']['end'] = self._io.pos()
                self._debug['value']['start'] = self._io.pos()
                self.value = self._io.read_bytes(self.len_value)
                self._debug['value']['end'] = self._io.pos()


        @property
        def type_list_and_reference_lists(self):
            """The resource map's resource type list, followed by the resource reference list area."""
            if hasattr(self, '_m_type_list_and_reference_lists'):
                return self._m_type_list_and_reference_lists if hasattr(self, '_m_type_list_and_reference_lists') else None

            _pos = self._io.pos()
            self._io.seek(self.ofs_type_list)
            self._debug['_m_type_list_and_reference_lists']['start'] = self._io.pos()
            self._raw__m_type_list_and_reference_lists = self._io.read_bytes((self.ofs_names - self.ofs_type_list))
            _io__raw__m_type_list_and_reference_lists = KaitaiStream(BytesIO(self._raw__m_type_list_and_reference_lists))
            self._m_type_list_and_reference_lists = ResourceFork.ResourceMap.TypeListAndReferenceLists(_io__raw__m_type_list_and_reference_lists, self, self._root)
            self._m_type_list_and_reference_lists._read()
            self._debug['_m_type_list_and_reference_lists']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_type_list_and_reference_lists if hasattr(self, '_m_type_list_and_reference_lists') else None

        @property
        def names_with_io(self):
            """Use `names` instead,
            unless you need access to this instance's `_io`.
            """
            if hasattr(self, '_m_names_with_io'):
                return self._m_names_with_io if hasattr(self, '_m_names_with_io') else None

            _pos = self._io.pos()
            self._io.seek(self.ofs_names)
            self._debug['_m_names_with_io']['start'] = self._io.pos()
            self._raw__m_names_with_io = self._io.read_bytes_full()
            _io__raw__m_names_with_io = KaitaiStream(BytesIO(self._raw__m_names_with_io))
            self._m_names_with_io = bytes_with_io.BytesWithIo(_io__raw__m_names_with_io)
            self._m_names_with_io._read()
            self._debug['_m_names_with_io']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_names_with_io if hasattr(self, '_m_names_with_io') else None

        @property
        def names(self):
            """Storage area for the names of all resources."""
            if hasattr(self, '_m_names'):
                return self._m_names if hasattr(self, '_m_names') else None

            self._m_names = self.names_with_io.data
            return self._m_names if hasattr(self, '_m_names') else None


    @property
    def data_blocks_with_io(self):
        """Use `data_blocks` instead,
        unless you need access to this instance's `_io`.
        """
        if hasattr(self, '_m_data_blocks_with_io'):
            return self._m_data_blocks_with_io if hasattr(self, '_m_data_blocks_with_io') else None

        _pos = self._io.pos()
        self._io.seek(self.header.ofs_data_blocks)
        self._debug['_m_data_blocks_with_io']['start'] = self._io.pos()
        self._raw__m_data_blocks_with_io = self._io.read_bytes(self.header.len_data_blocks)
        _io__raw__m_data_blocks_with_io = KaitaiStream(BytesIO(self._raw__m_data_blocks_with_io))
        self._m_data_blocks_with_io = bytes_with_io.BytesWithIo(_io__raw__m_data_blocks_with_io)
        self._m_data_blocks_with_io._read()
        self._debug['_m_data_blocks_with_io']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_data_blocks_with_io if hasattr(self, '_m_data_blocks_with_io') else None

    @property
    def data_blocks(self):
        """Storage area for the data blocks of all resources.
        
        These data blocks are not required to appear in any particular order,
        and there may be unused space between and around them.
        
        In practice,
        the data blocks in newly created resource files are usually contiguous.
        When existing resources are shortened,
        the Mac OS resource manager leaves unused space where the now removed resource data was,
        as this is quicker than moving the following resource data into the newly freed space.
        Such unused space may be cleaned up later when the resource manager "compacts" the resource file,
        which happens when resources are removed entirely,
        or when resources are added or grown so that more space is needed in the data area.
        """
        if hasattr(self, '_m_data_blocks'):
            return self._m_data_blocks if hasattr(self, '_m_data_blocks') else None

        self._m_data_blocks = self.data_blocks_with_io.data
        return self._m_data_blocks if hasattr(self, '_m_data_blocks') else None

    @property
    def resource_map(self):
        """The resource file's resource map."""
        if hasattr(self, '_m_resource_map'):
            return self._m_resource_map if hasattr(self, '_m_resource_map') else None

        _pos = self._io.pos()
        self._io.seek(self.header.ofs_resource_map)
        self._debug['_m_resource_map']['start'] = self._io.pos()
        self._raw__m_resource_map = self._io.read_bytes(self.header.len_resource_map)
        _io__raw__m_resource_map = KaitaiStream(BytesIO(self._raw__m_resource_map))
        self._m_resource_map = ResourceFork.ResourceMap(_io__raw__m_resource_map, self, self._root)
        self._m_resource_map._read()
        self._debug['_m_resource_map']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_resource_map if hasattr(self, '_m_resource_map') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class BlenderBlend(KaitaiStruct):
    """Blender is an open source suite for 3D modelling, sculpting,
    animation, compositing, rendering, preparation of assets for its own
    game engine and exporting to others, etc. `.blend` is its own binary
    format that saves whole state of suite: current scene, animations,
    all software settings, extensions, etc.
    
    Internally, .blend format is a hybrid semi-self-descriptive
    format. On top level, it contains a simple header and a sequence of
    file blocks, which more or less follow typical [TLV
    pattern](https://en.wikipedia.org/wiki/Type-length-value). Pre-last
    block would be a structure with code `DNA1`, which is a essentially
    a machine-readable schema of all other structures used in this file.
    """

    class PtrSize(Enum):
        bits_64 = 45
        bits_32 = 95

    class Endian(Enum):
        be = 86
        le = 118
    SEQ_FIELDS = ["hdr", "blocks"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['hdr']['start'] = self._io.pos()
        self.hdr = BlenderBlend.Header(self._io, self, self._root)
        self.hdr._read()
        self._debug['hdr']['end'] = self._io.pos()
        self._debug['blocks']['start'] = self._io.pos()
        self.blocks = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['blocks']:
                self._debug['blocks']['arr'] = []
            self._debug['blocks']['arr'].append({'start': self._io.pos()})
            _t_blocks = BlenderBlend.FileBlock(self._io, self, self._root)
            _t_blocks._read()
            self.blocks.append(_t_blocks)
            self._debug['blocks']['arr'][len(self.blocks) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['blocks']['end'] = self._io.pos()

    class DnaStruct(KaitaiStruct):
        """DNA struct contains a `type` (type name), which is specified as
        an index in types table, and sequence of fields.
        """
        SEQ_FIELDS = ["idx_type", "num_fields", "fields"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['idx_type']['start'] = self._io.pos()
            self.idx_type = self._io.read_u2le()
            self._debug['idx_type']['end'] = self._io.pos()
            self._debug['num_fields']['start'] = self._io.pos()
            self.num_fields = self._io.read_u2le()
            self._debug['num_fields']['end'] = self._io.pos()
            self._debug['fields']['start'] = self._io.pos()
            self.fields = [None] * (self.num_fields)
            for i in range(self.num_fields):
                if not 'arr' in self._debug['fields']:
                    self._debug['fields']['arr'] = []
                self._debug['fields']['arr'].append({'start': self._io.pos()})
                _t_fields = BlenderBlend.DnaField(self._io, self, self._root)
                _t_fields._read()
                self.fields[i] = _t_fields
                self._debug['fields']['arr'][i]['end'] = self._io.pos()

            self._debug['fields']['end'] = self._io.pos()

        @property
        def type(self):
            if hasattr(self, '_m_type'):
                return self._m_type if hasattr(self, '_m_type') else None

            self._m_type = self._parent.types[self.idx_type]
            return self._m_type if hasattr(self, '_m_type') else None


    class FileBlock(KaitaiStruct):
        SEQ_FIELDS = ["code", "len_body", "mem_addr", "sdna_index", "count", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = (self._io.read_bytes(4)).decode(u"ASCII")
            self._debug['code']['end'] = self._io.pos()
            self._debug['len_body']['start'] = self._io.pos()
            self.len_body = self._io.read_u4le()
            self._debug['len_body']['end'] = self._io.pos()
            self._debug['mem_addr']['start'] = self._io.pos()
            self.mem_addr = self._io.read_bytes(self._root.hdr.psize)
            self._debug['mem_addr']['end'] = self._io.pos()
            self._debug['sdna_index']['start'] = self._io.pos()
            self.sdna_index = self._io.read_u4le()
            self._debug['sdna_index']['end'] = self._io.pos()
            self._debug['count']['start'] = self._io.pos()
            self.count = self._io.read_u4le()
            self._debug['count']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.code
            if _on == u"DNA1":
                self._raw_body = self._io.read_bytes(self.len_body)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = BlenderBlend.Dna1Body(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.len_body)
            self._debug['body']['end'] = self._io.pos()

        @property
        def sdna_struct(self):
            if hasattr(self, '_m_sdna_struct'):
                return self._m_sdna_struct if hasattr(self, '_m_sdna_struct') else None

            if self.sdna_index != 0:
                self._m_sdna_struct = self._root.sdna_structs[self.sdna_index]

            return self._m_sdna_struct if hasattr(self, '_m_sdna_struct') else None


    class Dna1Body(KaitaiStruct):
        """DNA1, also known as "Structure DNA", is a special block in
        .blend file, which contains machine-readable specifications of
        all other structures used in this .blend file.
        
        Effectively, this block contains:
        
        * a sequence of "names" (strings which represent field names)
        * a sequence of "types" (strings which represent type name)
        * a sequence of "type lengths"
        * a sequence of "structs" (which describe contents of every
          structure, referring to types and names by index)
        
        .. seealso::
           Source - https://en.blender.org/index.php/Dev:Source/Architecture/File_Format#Structure_DNA
        """
        SEQ_FIELDS = ["id", "name_magic", "num_names", "names", "padding_1", "type_magic", "num_types", "types", "padding_2", "tlen_magic", "lengths", "padding_3", "strc_magic", "num_structs", "structs"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['id']['start'] = self._io.pos()
            self.id = self._io.read_bytes(4)
            self._debug['id']['end'] = self._io.pos()
            if not self.id == b"\x53\x44\x4E\x41":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\x44\x4E\x41", self.id, self._io, u"/types/dna1_body/seq/0")
            self._debug['name_magic']['start'] = self._io.pos()
            self.name_magic = self._io.read_bytes(4)
            self._debug['name_magic']['end'] = self._io.pos()
            if not self.name_magic == b"\x4E\x41\x4D\x45":
                raise kaitaistruct.ValidationNotEqualError(b"\x4E\x41\x4D\x45", self.name_magic, self._io, u"/types/dna1_body/seq/1")
            self._debug['num_names']['start'] = self._io.pos()
            self.num_names = self._io.read_u4le()
            self._debug['num_names']['end'] = self._io.pos()
            self._debug['names']['start'] = self._io.pos()
            self.names = [None] * (self.num_names)
            for i in range(self.num_names):
                if not 'arr' in self._debug['names']:
                    self._debug['names']['arr'] = []
                self._debug['names']['arr'].append({'start': self._io.pos()})
                self.names[i] = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
                self._debug['names']['arr'][i]['end'] = self._io.pos()

            self._debug['names']['end'] = self._io.pos()
            self._debug['padding_1']['start'] = self._io.pos()
            self.padding_1 = self._io.read_bytes(((4 - self._io.pos()) % 4))
            self._debug['padding_1']['end'] = self._io.pos()
            self._debug['type_magic']['start'] = self._io.pos()
            self.type_magic = self._io.read_bytes(4)
            self._debug['type_magic']['end'] = self._io.pos()
            if not self.type_magic == b"\x54\x59\x50\x45":
                raise kaitaistruct.ValidationNotEqualError(b"\x54\x59\x50\x45", self.type_magic, self._io, u"/types/dna1_body/seq/5")
            self._debug['num_types']['start'] = self._io.pos()
            self.num_types = self._io.read_u4le()
            self._debug['num_types']['end'] = self._io.pos()
            self._debug['types']['start'] = self._io.pos()
            self.types = [None] * (self.num_types)
            for i in range(self.num_types):
                if not 'arr' in self._debug['types']:
                    self._debug['types']['arr'] = []
                self._debug['types']['arr'].append({'start': self._io.pos()})
                self.types[i] = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
                self._debug['types']['arr'][i]['end'] = self._io.pos()

            self._debug['types']['end'] = self._io.pos()
            self._debug['padding_2']['start'] = self._io.pos()
            self.padding_2 = self._io.read_bytes(((4 - self._io.pos()) % 4))
            self._debug['padding_2']['end'] = self._io.pos()
            self._debug['tlen_magic']['start'] = self._io.pos()
            self.tlen_magic = self._io.read_bytes(4)
            self._debug['tlen_magic']['end'] = self._io.pos()
            if not self.tlen_magic == b"\x54\x4C\x45\x4E":
                raise kaitaistruct.ValidationNotEqualError(b"\x54\x4C\x45\x4E", self.tlen_magic, self._io, u"/types/dna1_body/seq/9")
            self._debug['lengths']['start'] = self._io.pos()
            self.lengths = [None] * (self.num_types)
            for i in range(self.num_types):
                if not 'arr' in self._debug['lengths']:
                    self._debug['lengths']['arr'] = []
                self._debug['lengths']['arr'].append({'start': self._io.pos()})
                self.lengths[i] = self._io.read_u2le()
                self._debug['lengths']['arr'][i]['end'] = self._io.pos()

            self._debug['lengths']['end'] = self._io.pos()
            self._debug['padding_3']['start'] = self._io.pos()
            self.padding_3 = self._io.read_bytes(((4 - self._io.pos()) % 4))
            self._debug['padding_3']['end'] = self._io.pos()
            self._debug['strc_magic']['start'] = self._io.pos()
            self.strc_magic = self._io.read_bytes(4)
            self._debug['strc_magic']['end'] = self._io.pos()
            if not self.strc_magic == b"\x53\x54\x52\x43":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\x54\x52\x43", self.strc_magic, self._io, u"/types/dna1_body/seq/12")
            self._debug['num_structs']['start'] = self._io.pos()
            self.num_structs = self._io.read_u4le()
            self._debug['num_structs']['end'] = self._io.pos()
            self._debug['structs']['start'] = self._io.pos()
            self.structs = [None] * (self.num_structs)
            for i in range(self.num_structs):
                if not 'arr' in self._debug['structs']:
                    self._debug['structs']['arr'] = []
                self._debug['structs']['arr'].append({'start': self._io.pos()})
                _t_structs = BlenderBlend.DnaStruct(self._io, self, self._root)
                _t_structs._read()
                self.structs[i] = _t_structs
                self._debug['structs']['arr'][i]['end'] = self._io.pos()

            self._debug['structs']['end'] = self._io.pos()


    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "ptr_size_id", "endian", "version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(7)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x42\x4C\x45\x4E\x44\x45\x52":
                raise kaitaistruct.ValidationNotEqualError(b"\x42\x4C\x45\x4E\x44\x45\x52", self.magic, self._io, u"/types/header/seq/0")
            self._debug['ptr_size_id']['start'] = self._io.pos()
            self.ptr_size_id = KaitaiStream.resolve_enum(BlenderBlend.PtrSize, self._io.read_u1())
            self._debug['ptr_size_id']['end'] = self._io.pos()
            self._debug['endian']['start'] = self._io.pos()
            self.endian = KaitaiStream.resolve_enum(BlenderBlend.Endian, self._io.read_u1())
            self._debug['endian']['end'] = self._io.pos()
            self._debug['version']['start'] = self._io.pos()
            self.version = (self._io.read_bytes(3)).decode(u"ASCII")
            self._debug['version']['end'] = self._io.pos()

        @property
        def psize(self):
            """Number of bytes that a pointer occupies."""
            if hasattr(self, '_m_psize'):
                return self._m_psize if hasattr(self, '_m_psize') else None

            self._m_psize = (8 if self.ptr_size_id == BlenderBlend.PtrSize.bits_64 else 4)
            return self._m_psize if hasattr(self, '_m_psize') else None


    class DnaField(KaitaiStruct):
        SEQ_FIELDS = ["idx_type", "idx_name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['idx_type']['start'] = self._io.pos()
            self.idx_type = self._io.read_u2le()
            self._debug['idx_type']['end'] = self._io.pos()
            self._debug['idx_name']['start'] = self._io.pos()
            self.idx_name = self._io.read_u2le()
            self._debug['idx_name']['end'] = self._io.pos()

        @property
        def type(self):
            if hasattr(self, '_m_type'):
                return self._m_type if hasattr(self, '_m_type') else None

            self._m_type = self._parent._parent.types[self.idx_type]
            return self._m_type if hasattr(self, '_m_type') else None

        @property
        def name(self):
            if hasattr(self, '_m_name'):
                return self._m_name if hasattr(self, '_m_name') else None

            self._m_name = self._parent._parent.names[self.idx_name]
            return self._m_name if hasattr(self, '_m_name') else None


    @property
    def sdna_structs(self):
        if hasattr(self, '_m_sdna_structs'):
            return self._m_sdna_structs if hasattr(self, '_m_sdna_structs') else None

        self._m_sdna_structs = self.blocks[(len(self.blocks) - 2)].body.structs
        return self._m_sdna_structs if hasattr(self, '_m_sdna_structs') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Vp8Ivf(KaitaiStruct):
    """IVF is a simple container format for raw VP8 data, which is an open
    and royalty-free video compression format, currently developed by
    Google.
    
    Test .ivf files are available at https://chromium.googlesource.com/webm/vp8-test-vectors
    
    .. seealso::
       Source - https://wiki.multimedia.cx/index.php/IVF
    """
    SEQ_FIELDS = ["magic1", "version", "len_header", "codec", "width", "height", "framerate", "timescale", "num_frames", "unused", "image_data"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic1']['start'] = self._io.pos()
        self.magic1 = self._io.read_bytes(4)
        self._debug['magic1']['end'] = self._io.pos()
        if not self.magic1 == b"\x44\x4B\x49\x46":
            raise kaitaistruct.ValidationNotEqualError(b"\x44\x4B\x49\x46", self.magic1, self._io, u"/seq/0")
        self._debug['version']['start'] = self._io.pos()
        self.version = self._io.read_u2le()
        self._debug['version']['end'] = self._io.pos()
        self._debug['len_header']['start'] = self._io.pos()
        self.len_header = self._io.read_u2le()
        self._debug['len_header']['end'] = self._io.pos()
        self._debug['codec']['start'] = self._io.pos()
        self.codec = self._io.read_bytes(4)
        self._debug['codec']['end'] = self._io.pos()
        if not self.codec == b"\x56\x50\x38\x30":
            raise kaitaistruct.ValidationNotEqualError(b"\x56\x50\x38\x30", self.codec, self._io, u"/seq/3")
        self._debug['width']['start'] = self._io.pos()
        self.width = self._io.read_u2le()
        self._debug['width']['end'] = self._io.pos()
        self._debug['height']['start'] = self._io.pos()
        self.height = self._io.read_u2le()
        self._debug['height']['end'] = self._io.pos()
        self._debug['framerate']['start'] = self._io.pos()
        self.framerate = self._io.read_u4le()
        self._debug['framerate']['end'] = self._io.pos()
        self._debug['timescale']['start'] = self._io.pos()
        self.timescale = self._io.read_u4le()
        self._debug['timescale']['end'] = self._io.pos()
        self._debug['num_frames']['start'] = self._io.pos()
        self.num_frames = self._io.read_u4le()
        self._debug['num_frames']['end'] = self._io.pos()
        self._debug['unused']['start'] = self._io.pos()
        self.unused = self._io.read_u4le()
        self._debug['unused']['end'] = self._io.pos()
        self._debug['image_data']['start'] = self._io.pos()
        self.image_data = [None] * (self.num_frames)
        for i in range(self.num_frames):
            if not 'arr' in self._debug['image_data']:
                self._debug['image_data']['arr'] = []
            self._debug['image_data']['arr'].append({'start': self._io.pos()})
            _t_image_data = Vp8Ivf.Blocks(self._io, self, self._root)
            _t_image_data._read()
            self.image_data[i] = _t_image_data
            self._debug['image_data']['arr'][i]['end'] = self._io.pos()

        self._debug['image_data']['end'] = self._io.pos()

    class Blocks(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = Vp8Ivf.Block(self._io, self, self._root)
            self.entries._read()
            self._debug['entries']['end'] = self._io.pos()


    class Block(KaitaiStruct):
        SEQ_FIELDS = ["len_frame", "timestamp", "framedata"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_frame']['start'] = self._io.pos()
            self.len_frame = self._io.read_u4le()
            self._debug['len_frame']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u8le()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['framedata']['start'] = self._io.pos()
            self.framedata = self._io.read_bytes(self.len_frame)
            self._debug['framedata']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Hccapx(KaitaiStruct):
    """Native format of Hashcat password "recovery" utility
    
    .. seealso::
       Source - https://hashcat.net/wiki/doku.php?id=hccapx
    """
    SEQ_FIELDS = ["records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['records']['start'] = self._io.pos()
        self.records = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            _t_records = Hccapx.HccapxRecord(self._io, self, self._root)
            _t_records._read()
            self.records.append(_t_records)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['records']['end'] = self._io.pos()

    class HccapxRecord(KaitaiStruct):
        SEQ_FIELDS = ["magic", "version", "ignore_replay_counter", "message_pair", "len_essid", "essid", "padding1", "keyver", "keymic", "mac_ap", "nonce_ap", "mac_station", "nonce_station", "len_eapol", "eapol", "padding2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x48\x43\x50\x58":
                raise kaitaistruct.ValidationNotEqualError(b"\x48\x43\x50\x58", self.magic, self._io, u"/types/hccapx_record/seq/0")
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u4le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['ignore_replay_counter']['start'] = self._io.pos()
            self.ignore_replay_counter = self._io.read_bits_int_be(1) != 0
            self._debug['ignore_replay_counter']['end'] = self._io.pos()
            self._debug['message_pair']['start'] = self._io.pos()
            self.message_pair = self._io.read_bits_int_be(7)
            self._debug['message_pair']['end'] = self._io.pos()
            self._io.align_to_byte()
            self._debug['len_essid']['start'] = self._io.pos()
            self.len_essid = self._io.read_u1()
            self._debug['len_essid']['end'] = self._io.pos()
            self._debug['essid']['start'] = self._io.pos()
            self.essid = self._io.read_bytes(self.len_essid)
            self._debug['essid']['end'] = self._io.pos()
            self._debug['padding1']['start'] = self._io.pos()
            self.padding1 = self._io.read_bytes((32 - self.len_essid))
            self._debug['padding1']['end'] = self._io.pos()
            self._debug['keyver']['start'] = self._io.pos()
            self.keyver = self._io.read_u1()
            self._debug['keyver']['end'] = self._io.pos()
            self._debug['keymic']['start'] = self._io.pos()
            self.keymic = self._io.read_bytes(16)
            self._debug['keymic']['end'] = self._io.pos()
            self._debug['mac_ap']['start'] = self._io.pos()
            self.mac_ap = self._io.read_bytes(6)
            self._debug['mac_ap']['end'] = self._io.pos()
            self._debug['nonce_ap']['start'] = self._io.pos()
            self.nonce_ap = self._io.read_bytes(32)
            self._debug['nonce_ap']['end'] = self._io.pos()
            self._debug['mac_station']['start'] = self._io.pos()
            self.mac_station = self._io.read_bytes(6)
            self._debug['mac_station']['end'] = self._io.pos()
            self._debug['nonce_station']['start'] = self._io.pos()
            self.nonce_station = self._io.read_bytes(32)
            self._debug['nonce_station']['end'] = self._io.pos()
            self._debug['len_eapol']['start'] = self._io.pos()
            self.len_eapol = self._io.read_u2le()
            self._debug['len_eapol']['end'] = self._io.pos()
            self._debug['eapol']['start'] = self._io.pos()
            self.eapol = self._io.read_bytes(self.len_eapol)
            self._debug['eapol']['end'] = self._io.pos()
            self._debug['padding2']['start'] = self._io.pos()
            self.padding2 = self._io.read_bytes((256 - self.len_eapol))
            self._debug['padding2']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Id3v11(KaitaiStruct):
    """ID3v1.1 tag is a method to store simple metadata in .mp3 files. The
    tag is appended to the end of file and spans exactly 128 bytes.
    
    This type is supposed to be used on full .mp3 files, seeking to
    proper position automatically. If you're interesting in parsing only
    the tag itself, please use `id3v1_1::id3_v1_1_tag` subtype.
    
    .. seealso::
       Source - https://id3.org/ID3v1
    """
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class Id3V11Tag(KaitaiStruct):
        """ID3v1.1 tag itself, a fixed size 128 byte structure. Contains
        several metadata fields as fixed-size strings.
        
        Note that string encoding is not specified by standard, so real
        encoding used would vary a lot from one implementation to
        another. Most Windows-based applications tend to use "ANSI"
        (i.e. locale-dependent encoding, usually one byte per
        character). Some embedded applications allow selection of
        charset.
        """

        class GenreEnum(Enum):
            blues = 0
            classic_rock = 1
            country = 2
            dance = 3
            disco = 4
            funk = 5
            grunge = 6
            hip_hop = 7
            jazz = 8
            metal = 9
            new_age = 10
            oldies = 11
            other = 12
            pop = 13
            rnb = 14
            rap = 15
            reggae = 16
            rock = 17
            techno = 18
            industrial = 19
            alternative = 20
            ska = 21
            death_metal = 22
            pranks = 23
            soundtrack = 24
            euro_techno = 25
            ambient = 26
            trip_hop = 27
            vocal = 28
            jazz_funk = 29
            fusion = 30
            trance = 31
            classical = 32
            instrumental = 33
            acid = 34
            house = 35
            game = 36
            sound_clip = 37
            gospel = 38
            noise = 39
            alternrock = 40
            bass = 41
            soul = 42
            punk = 43
            space = 44
            meditative = 45
            instrumental_pop = 46
            instrumental_rock = 47
            ethnic = 48
            gothic = 49
            darkwave = 50
            techno_industrial = 51
            electronic = 52
            pop_folk = 53
            eurodance = 54
            dream = 55
            southern_rock = 56
            comedy = 57
            cult = 58
            gangsta = 59
            top_40 = 60
            christian_rap = 61
            pop_funk = 62
            jungle = 63
            native_american = 64
            cabaret = 65
            new_wave = 66
            psychadelic = 67
            rave = 68
            showtunes = 69
            trailer = 70
            lo_fi = 71
            tribal = 72
            acid_punk = 73
            acid_jazz = 74
            polka = 75
            retro = 76
            musical = 77
            rock_n_roll = 78
            hard_rock = 79
            folk = 80
            folk_rock = 81
            national_folk = 82
            swing = 83
            fast_fusion = 84
            bebob = 85
            latin = 86
            revival = 87
            celtic = 88
            bluegrass = 89
            avantgarde = 90
            gothic_rock = 91
            progressive_rock = 92
            psychedelic_rock = 93
            symphonic_rock = 94
            slow_rock = 95
            big_band = 96
            chorus = 97
            easy_listening = 98
            acoustic = 99
            humour = 100
            speech = 101
            chanson = 102
            opera = 103
            chamber_music = 104
            sonata = 105
            symphony = 106
            booty_bass = 107
            primus = 108
            porn_groove = 109
            satire = 110
            slow_jam = 111
            club = 112
            tango = 113
            samba = 114
            folklore = 115
            ballad = 116
            power_ballad = 117
            rhythmic_soul = 118
            freestyle = 119
            duet = 120
            punk_rock = 121
            drum_solo = 122
            a_capella = 123
            euro_house = 124
            dance_hall = 125
        SEQ_FIELDS = ["magic", "title", "artist", "album", "year", "comment", "genre"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(3)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x54\x41\x47":
                raise kaitaistruct.ValidationNotEqualError(b"\x54\x41\x47", self.magic, self._io, u"/types/id3_v1_1_tag/seq/0")
            self._debug['title']['start'] = self._io.pos()
            self.title = self._io.read_bytes(30)
            self._debug['title']['end'] = self._io.pos()
            self._debug['artist']['start'] = self._io.pos()
            self.artist = self._io.read_bytes(30)
            self._debug['artist']['end'] = self._io.pos()
            self._debug['album']['start'] = self._io.pos()
            self.album = self._io.read_bytes(30)
            self._debug['album']['end'] = self._io.pos()
            self._debug['year']['start'] = self._io.pos()
            self.year = (self._io.read_bytes(4)).decode(u"ASCII")
            self._debug['year']['end'] = self._io.pos()
            self._debug['comment']['start'] = self._io.pos()
            self.comment = self._io.read_bytes(30)
            self._debug['comment']['end'] = self._io.pos()
            self._debug['genre']['start'] = self._io.pos()
            self.genre = KaitaiStream.resolve_enum(Id3v11.Id3V11Tag.GenreEnum, self._io.read_u1())
            self._debug['genre']['end'] = self._io.pos()


    @property
    def id3v1_tag(self):
        if hasattr(self, '_m_id3v1_tag'):
            return self._m_id3v1_tag if hasattr(self, '_m_id3v1_tag') else None

        _pos = self._io.pos()
        self._io.seek((self._io.size() - 128))
        self._debug['_m_id3v1_tag']['start'] = self._io.pos()
        self._m_id3v1_tag = Id3v11.Id3V11Tag(self._io, self, self._root)
        self._m_id3v1_tag._read()
        self._debug['_m_id3v1_tag']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_id3v1_tag if hasattr(self, '_m_id3v1_tag') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class WindowsShellItems(KaitaiStruct):
    """Windows Shell Items (AKA "shellbags") is an undocumented set of
    structures used internally within Windows to identify paths in
    Windows Folder Hierarchy. It is widely used in Windows Shell (and
    most visible in File Explorer), both as in-memory and in-file
    structures. Some formats embed them, namely:
    
    * Windows Shell link files (.lnk) Windows registry
    * Windows registry "ShellBags" keys
    
    The format is mostly undocumented, and is known to vary between
    various Windows versions.
    
    .. seealso::
       Source - https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc
    """
    SEQ_FIELDS = ["items"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['items']['start'] = self._io.pos()
        self.items = []
        i = 0
        while True:
            if not 'arr' in self._debug['items']:
                self._debug['items']['arr'] = []
            self._debug['items']['arr'].append({'start': self._io.pos()})
            _t_items = WindowsShellItems.ShellItem(self._io, self, self._root)
            _t_items._read()
            _ = _t_items
            self.items.append(_)
            self._debug['items']['arr'][len(self.items) - 1]['end'] = self._io.pos()
            if _.len_data == 0:
                break
            i += 1
        self._debug['items']['end'] = self._io.pos()

    class ShellItemData(KaitaiStruct):
        SEQ_FIELDS = ["code", "body1", "body2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = self._io.read_u1()
            self._debug['code']['end'] = self._io.pos()
            self._debug['body1']['start'] = self._io.pos()
            _on = self.code
            if _on == 31:
                self.body1 = WindowsShellItems.RootFolderBody(self._io, self, self._root)
                self.body1._read()
            self._debug['body1']['end'] = self._io.pos()
            self._debug['body2']['start'] = self._io.pos()
            _on = (self.code & 112)
            if _on == 32:
                self.body2 = WindowsShellItems.VolumeBody(self._io, self, self._root)
                self.body2._read()
            elif _on == 48:
                self.body2 = WindowsShellItems.FileEntryBody(self._io, self, self._root)
                self.body2._read()
            self._debug['body2']['end'] = self._io.pos()


    class ShellItem(KaitaiStruct):
        """
        .. seealso::
           Section 2.2.2 - https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-SHLLINK/[MS-SHLLINK].pdf
        """
        SEQ_FIELDS = ["len_data", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u2le()
            self._debug['len_data']['end'] = self._io.pos()
            if self.len_data >= 2:
                self._debug['data']['start'] = self._io.pos()
                self._raw_data = self._io.read_bytes((self.len_data - 2))
                _io__raw_data = KaitaiStream(BytesIO(self._raw_data))
                self.data = WindowsShellItems.ShellItemData(_io__raw_data, self, self._root)
                self.data._read()
                self._debug['data']['end'] = self._io.pos()



    class RootFolderBody(KaitaiStruct):
        """
        .. seealso::
           Source - https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#32-root-folder-shell-item
        """
        SEQ_FIELDS = ["sort_index", "shell_folder_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['sort_index']['start'] = self._io.pos()
            self.sort_index = self._io.read_u1()
            self._debug['sort_index']['end'] = self._io.pos()
            self._debug['shell_folder_id']['start'] = self._io.pos()
            self.shell_folder_id = self._io.read_bytes(16)
            self._debug['shell_folder_id']['end'] = self._io.pos()


    class VolumeBody(KaitaiStruct):
        """
        .. seealso::
           Source - https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#33-volume-shell-item
        """
        SEQ_FIELDS = ["flags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u1()
            self._debug['flags']['end'] = self._io.pos()


    class FileEntryBody(KaitaiStruct):
        """
        .. seealso::
           Source - https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#34-file-entry-shell-item
        """
        SEQ_FIELDS = ["_unnamed0", "file_size", "last_mod_time", "file_attrs"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['_unnamed0']['start'] = self._io.pos()
            self._unnamed0 = self._io.read_u1()
            self._debug['_unnamed0']['end'] = self._io.pos()
            self._debug['file_size']['start'] = self._io.pos()
            self.file_size = self._io.read_u4le()
            self._debug['file_size']['end'] = self._io.pos()
            self._debug['last_mod_time']['start'] = self._io.pos()
            self.last_mod_time = self._io.read_u4le()
            self._debug['last_mod_time']['end'] = self._io.pos()
            self._debug['file_attrs']['start'] = self._io.pos()
            self.file_attrs = self._io.read_u2le()
            self._debug['file_attrs']['end'] = self._io.pos()

        @property
        def is_dir(self):
            if hasattr(self, '_m_is_dir'):
                return self._m_is_dir if hasattr(self, '_m_is_dir') else None

            self._m_is_dir = (self._parent.code & 1) != 0
            return self._m_is_dir if hasattr(self, '_m_is_dir') else None

        @property
        def is_file(self):
            if hasattr(self, '_m_is_file'):
                return self._m_is_file if hasattr(self, '_m_is_file') else None

            self._m_is_file = (self._parent.code & 2) != 0
            return self._m_is_file if hasattr(self, '_m_is_file') else None






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AndroidDto(KaitaiStruct):
    """Format for Android DTB/DTBO partitions. It's kind of archive with
    dtb/dtbo files. Used only when there is a separate unique partition
    (dtb, dtbo) on an android device to organize device tree files.
    The format consists of a header with info about size and number
    of device tree entries and the entries themselves. This format
    description could be used to extract device tree entries from a
    partition images and decompile them with dtc (device tree compiler).
    
    .. seealso::
       Source - https://source.android.com/devices/architecture/dto/partitions
    
    
    .. seealso::
       Source - https://android.googlesource.com/platform/system/libufdt/+/refs/tags/android-10.0.0_r47
    """
    SEQ_FIELDS = ["header", "entries"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = AndroidDto.DtTableHeader(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['entries']['start'] = self._io.pos()
        self.entries = [None] * (self.header.dt_entry_count)
        for i in range(self.header.dt_entry_count):
            if not 'arr' in self._debug['entries']:
                self._debug['entries']['arr'] = []
            self._debug['entries']['arr'].append({'start': self._io.pos()})
            _t_entries = AndroidDto.DtTableEntry(self._io, self, self._root)
            _t_entries._read()
            self.entries[i] = _t_entries
            self._debug['entries']['arr'][i]['end'] = self._io.pos()

        self._debug['entries']['end'] = self._io.pos()

    class DtTableHeader(KaitaiStruct):
        SEQ_FIELDS = ["magic", "total_size", "header_size", "dt_entry_size", "dt_entry_count", "dt_entries_offset", "page_size", "version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\xD7\xB7\xAB\x1E":
                raise kaitaistruct.ValidationNotEqualError(b"\xD7\xB7\xAB\x1E", self.magic, self._io, u"/types/dt_table_header/seq/0")
            self._debug['total_size']['start'] = self._io.pos()
            self.total_size = self._io.read_u4be()
            self._debug['total_size']['end'] = self._io.pos()
            self._debug['header_size']['start'] = self._io.pos()
            self.header_size = self._io.read_u4be()
            self._debug['header_size']['end'] = self._io.pos()
            self._debug['dt_entry_size']['start'] = self._io.pos()
            self.dt_entry_size = self._io.read_u4be()
            self._debug['dt_entry_size']['end'] = self._io.pos()
            self._debug['dt_entry_count']['start'] = self._io.pos()
            self.dt_entry_count = self._io.read_u4be()
            self._debug['dt_entry_count']['end'] = self._io.pos()
            self._debug['dt_entries_offset']['start'] = self._io.pos()
            self.dt_entries_offset = self._io.read_u4be()
            self._debug['dt_entries_offset']['end'] = self._io.pos()
            self._debug['page_size']['start'] = self._io.pos()
            self.page_size = self._io.read_u4be()
            self._debug['page_size']['end'] = self._io.pos()
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u4be()
            self._debug['version']['end'] = self._io.pos()


    class DtTableEntry(KaitaiStruct):
        SEQ_FIELDS = ["dt_size", "dt_offset", "id", "rev", "custom"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['dt_size']['start'] = self._io.pos()
            self.dt_size = self._io.read_u4be()
            self._debug['dt_size']['end'] = self._io.pos()
            self._debug['dt_offset']['start'] = self._io.pos()
            self.dt_offset = self._io.read_u4be()
            self._debug['dt_offset']['end'] = self._io.pos()
            self._debug['id']['start'] = self._io.pos()
            self.id = self._io.read_u4be()
            self._debug['id']['end'] = self._io.pos()
            self._debug['rev']['start'] = self._io.pos()
            self.rev = self._io.read_u4be()
            self._debug['rev']['end'] = self._io.pos()
            self._debug['custom']['start'] = self._io.pos()
            self.custom = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['custom']:
                    self._debug['custom']['arr'] = []
                self._debug['custom']['arr'].append({'start': self._io.pos()})
                self.custom[i] = self._io.read_u4be()
                self._debug['custom']['arr'][i]['end'] = self._io.pos()

            self._debug['custom']['end'] = self._io.pos()

        @property
        def body(self):
            """DTB/DTBO file."""
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.dt_offset)
            self._debug['_m_body']['start'] = io.pos()
            self._m_body = io.read_bytes(self.dt_size)
            self._debug['_m_body']['end'] = io.pos()
            io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None






from collections import defaultdict
from typing import Dict, Iterable, List, Set, Tuple

from polyfile import logger

from .cfg import CFG

log = logger.getStatusLogger('PolyTracker')


class FunctionInfo:
    def __init__(self, name: str, cmp_bytes: Dict[str, List[int]], input_bytes: Dict[str, List[int]] = None, called_from: Iterable[str] = ()):
        self.name = name
        self.called_from = frozenset(called_from)
        self.cmp_bytes = cmp_bytes
        if input_bytes is None:
            self.input_bytes = cmp_bytes
        else:
            self.input_bytes = input_bytes

    @property
    def taint_sources(self) -> Set[str]:
        return self.input_bytes.keys()

    def __getitem__(self, input_source_name):
        return self.input_bytes[input_source_name]

    def __iter__(self):
        return self.taint_sources

    def items(self):
        return self.input_bytes.items()

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name!r}, cmp_bytes={self.cmp_bytes!r}, input_bytes={self.input_bytes!r}, called_from={self.called_from!r})"


class ProgramTrace:
    def __init__(self, polytracker_version: tuple, function_data: Iterable[FunctionInfo]):
        self.polytracker_version = polytracker_version
        self.functions: Dict[str, FunctionInfo] = {f.name: f for f in function_data}
        self._cfg = None

    @property
    def cfg(self) -> CFG:
        if self._cfg is not None:
            return self._cfg
        self._cfg = CFG(self)
        self._cfg.add_nodes_from(self.functions.values())
        for f in list(self.functions.values()):
            for caller in f.called_from:
                if caller not in self.functions:
                    info = FunctionInfo(caller, {})
                    self.functions[caller] = info
                    self._cfg.add_node(info)
                    self._cfg.add_edge(info, f)
                else:
                    self._cfg.add_edge(self.functions[caller], f)
        return self._cfg

    def __repr__(self):
        return f"{self.__class__.__name__}(polytracker_version={self.polytracker_version!r}, function_data={list(self.functions.values())!r})"


POLYTRACKER_JSON_FORMATS = []


def normalize_version(*version) -> Tuple[str]:
    version = tuple(str(v) for v in version)
    version = tuple(version) + ('0',) * (3 - len(version))
    version = tuple(version) + ('',) * (4 - len(version))
    return version


def polytracker_version(*version):
    def wrapper(func):
        POLYTRACKER_JSON_FORMATS.append((normalize_version(*version), func))
        POLYTRACKER_JSON_FORMATS.sort(reverse=True)
        return func
    return wrapper


def parse(polytracker_json_obj: dict) -> ProgramTrace:
    if 'version' in polytracker_json_obj:
        version = normalize_version(*polytracker_json_obj['version'].split('.'))
        if len(version) > 4:
            log.warn(f"Unexpectedly long PolyTracker version: {polytracker_json_obj['version']!r}")
        for i, (known_version, parser) in enumerate(POLYTRACKER_JSON_FORMATS):
            # POLYTRACKER_JSON_FORMATS is auto-sorted in decreasing order
            if version >= known_version:
                if i == 0 and version > known_version:
                    log.warn(f"PolyTracker version {polytracker_json_obj['version']!r} is newer than the latest supported by PolyMerge ({'.'.join(known_version)})")
                return parser(polytracker_json_obj)
        raise ValueError(f"Unsupported PolyTracker version {polytracker_json_obj['version']!r}")
    for function_name, function_data in polytracker_json_obj.items():
        if isinstance(function_data, dict) and 'called_from' in function_data:
            # this is the second version of the output format
            return parse_format_v2(polytracker_json_obj)
        else:
            return parse_format_v1(polytracker_json_obj)


@polytracker_version(0, 0, 1, '')
def parse_format_v1(polytracker_json_obj: dict) -> ProgramTrace:
    return ProgramTrace(
        polytracker_version=(0, 0, 1),
        function_data=[FunctionInfo(
            function_name,
            {None: taint_bytes}
        ) for function_name, taint_bytes in polytracker_json_obj.items()
        ]
    )


@polytracker_version(0, 0, 1, 'alpha2.1')
def parse_format_v2(polytracker_json_obj: dict) -> ProgramTrace:
    function_data = []
    for function_name, data in polytracker_json_obj.items():
        if 'input_bytes' not in data:
            if 'cmp_bytes' in data:
                input_bytes = data['cmp_bytes']
            else:
                input_bytes = {}
        else:
            input_bytes = data['input_bytes']
        if 'cmp_bytes' in data:
            cmp_bytes = data['cmp_bytes']
        else:
            cmp_bytes = input_bytes
        if 'called_from' in data:
            called_from = data['called_from']
        else:
            called_from = ()
        function_data.append(FunctionInfo(
            name=function_name,
            cmp_bytes=cmp_bytes,
            input_bytes=input_bytes,
            called_from=called_from
        ))
    return ProgramTrace(
        polytracker_version=(0, 0, 1, 'alpha2.1'),
        function_data=function_data
    )


@polytracker_version(1, 0, 1)
def parse_format_v3(polytracker_json_obj: dict) -> ProgramTrace:
    version = polytracker_json_obj['version'].split('.')
    function_data = []
    tainted_functions = set()
    for function_name, data in polytracker_json_obj['tainted_functions'].items():
        if 'input_bytes' not in data:
            if 'cmp_bytes' in data:
                input_bytes = data['cmp_bytes']
            else:
                input_bytes = {}
        else:
            input_bytes = data['input_bytes']
        if 'cmp_bytes' in data:
            cmp_bytes = data['cmp_bytes']
        else:
            cmp_bytes = input_bytes
        if function_name in polytracker_json_obj['runtime_cfg']:
            called_from = frozenset(polytracker_json_obj['runtime_cfg'][function_name])
        else:
            called_from = frozenset()
        function_data.append(FunctionInfo(
            name=function_name,
            cmp_bytes=cmp_bytes,
            input_bytes=input_bytes,
            called_from=called_from
        ))
        tainted_functions.add(function_name)
    # Add any additional functions from the CFG that didn't operate on tainted bytes
    for function_name in polytracker_json_obj['runtime_cfg'].keys() - tainted_functions:
        function_data.append(FunctionInfo(
            name=function_name,
            cmp_bytes={},
            called_from=polytracker_json_obj['runtime_cfg'][function_name]
        ))
    return ProgramTrace(
        polytracker_version=version,
        function_data=function_data
    )



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Lvm2(KaitaiStruct):
    """### Building a test file
    
    ```
    dd if=/dev/zero of=image.img bs=512 count=$(( 4 * 1024 * 2 ))
    sudo losetup /dev/loop1 image.img
    sudo pvcreate /dev/loop1
    sudo vgcreate vg_test /dev/loop1
    sudo lvcreate --name lv_test1 vg_test
    sudo losetup -d /dev/loop1
    ```
    
    .. seealso::
       Source - https://github.com/libyal/libvslvm/blob/master/documentation/Logical%20Volume%20Manager%20(LVM)%20format.asciidoc
    """
    SEQ_FIELDS = ["pv"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['pv']['start'] = self._io.pos()
        self.pv = Lvm2.PhysicalVolume(self._io, self, self._root)
        self.pv._read()
        self._debug['pv']['end'] = self._io.pos()

    class PhysicalVolume(KaitaiStruct):
        SEQ_FIELDS = ["empty_sector", "label"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['empty_sector']['start'] = self._io.pos()
            self.empty_sector = self._io.read_bytes(self._root.sector_size)
            self._debug['empty_sector']['end'] = self._io.pos()
            self._debug['label']['start'] = self._io.pos()
            self.label = Lvm2.PhysicalVolume.Label(self._io, self, self._root)
            self.label._read()
            self._debug['label']['end'] = self._io.pos()

        class Label(KaitaiStruct):
            SEQ_FIELDS = ["label_header", "volume_header"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['label_header']['start'] = self._io.pos()
                self.label_header = Lvm2.PhysicalVolume.Label.LabelHeader(self._io, self, self._root)
                self.label_header._read()
                self._debug['label_header']['end'] = self._io.pos()
                self._debug['volume_header']['start'] = self._io.pos()
                self.volume_header = Lvm2.PhysicalVolume.Label.VolumeHeader(self._io, self, self._root)
                self.volume_header._read()
                self._debug['volume_header']['end'] = self._io.pos()

            class LabelHeader(KaitaiStruct):
                SEQ_FIELDS = ["signature", "sector_number", "checksum", "label_header_"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['signature']['start'] = self._io.pos()
                    self.signature = self._io.read_bytes(8)
                    self._debug['signature']['end'] = self._io.pos()
                    if not self.signature == b"\x4C\x41\x42\x45\x4C\x4F\x4E\x45":
                        raise kaitaistruct.ValidationNotEqualError(b"\x4C\x41\x42\x45\x4C\x4F\x4E\x45", self.signature, self._io, u"/types/physical_volume/types/label/types/label_header/seq/0")
                    self._debug['sector_number']['start'] = self._io.pos()
                    self.sector_number = self._io.read_u8le()
                    self._debug['sector_number']['end'] = self._io.pos()
                    self._debug['checksum']['start'] = self._io.pos()
                    self.checksum = self._io.read_u4le()
                    self._debug['checksum']['end'] = self._io.pos()
                    self._debug['label_header_']['start'] = self._io.pos()
                    self.label_header_ = Lvm2.PhysicalVolume.Label.LabelHeader.LabelHeader(self._io, self, self._root)
                    self.label_header_._read()
                    self._debug['label_header_']['end'] = self._io.pos()

                class LabelHeader(KaitaiStruct):
                    SEQ_FIELDS = ["data_offset", "type_indicator"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['data_offset']['start'] = self._io.pos()
                        self.data_offset = self._io.read_u4le()
                        self._debug['data_offset']['end'] = self._io.pos()
                        self._debug['type_indicator']['start'] = self._io.pos()
                        self.type_indicator = self._io.read_bytes(8)
                        self._debug['type_indicator']['end'] = self._io.pos()
                        if not self.type_indicator == b"\x4C\x56\x4D\x32\x20\x30\x30\x31":
                            raise kaitaistruct.ValidationNotEqualError(b"\x4C\x56\x4D\x32\x20\x30\x30\x31", self.type_indicator, self._io, u"/types/physical_volume/types/label/types/label_header/types/label_header_/seq/1")



            class VolumeHeader(KaitaiStruct):
                SEQ_FIELDS = ["id", "size", "data_area_descriptors", "metadata_area_descriptors"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['id']['start'] = self._io.pos()
                    self.id = (self._io.read_bytes(32)).decode(u"ascii")
                    self._debug['id']['end'] = self._io.pos()
                    self._debug['size']['start'] = self._io.pos()
                    self.size = self._io.read_u8le()
                    self._debug['size']['end'] = self._io.pos()
                    self._debug['data_area_descriptors']['start'] = self._io.pos()
                    self.data_area_descriptors = []
                    i = 0
                    while True:
                        if not 'arr' in self._debug['data_area_descriptors']:
                            self._debug['data_area_descriptors']['arr'] = []
                        self._debug['data_area_descriptors']['arr'].append({'start': self._io.pos()})
                        _t_data_area_descriptors = Lvm2.PhysicalVolume.Label.VolumeHeader.DataAreaDescriptor(self._io, self, self._root)
                        _t_data_area_descriptors._read()
                        _ = _t_data_area_descriptors
                        self.data_area_descriptors.append(_)
                        self._debug['data_area_descriptors']['arr'][len(self.data_area_descriptors) - 1]['end'] = self._io.pos()
                        if  ((_.size != 0) and (_.offset != 0)) :
                            break
                        i += 1
                    self._debug['data_area_descriptors']['end'] = self._io.pos()
                    self._debug['metadata_area_descriptors']['start'] = self._io.pos()
                    self.metadata_area_descriptors = []
                    i = 0
                    while True:
                        if not 'arr' in self._debug['metadata_area_descriptors']:
                            self._debug['metadata_area_descriptors']['arr'] = []
                        self._debug['metadata_area_descriptors']['arr'].append({'start': self._io.pos()})
                        _t_metadata_area_descriptors = Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataAreaDescriptor(self._io, self, self._root)
                        _t_metadata_area_descriptors._read()
                        _ = _t_metadata_area_descriptors
                        self.metadata_area_descriptors.append(_)
                        self._debug['metadata_area_descriptors']['arr'][len(self.metadata_area_descriptors) - 1]['end'] = self._io.pos()
                        if  ((_.size != 0) and (_.offset != 0)) :
                            break
                        i += 1
                    self._debug['metadata_area_descriptors']['end'] = self._io.pos()

                class DataAreaDescriptor(KaitaiStruct):
                    SEQ_FIELDS = ["offset", "size"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['offset']['start'] = self._io.pos()
                        self.offset = self._io.read_u8le()
                        self._debug['offset']['end'] = self._io.pos()
                        self._debug['size']['start'] = self._io.pos()
                        self.size = self._io.read_u8le()
                        self._debug['size']['end'] = self._io.pos()

                    @property
                    def data(self):
                        if hasattr(self, '_m_data'):
                            return self._m_data if hasattr(self, '_m_data') else None

                        if self.size != 0:
                            _pos = self._io.pos()
                            self._io.seek(self.offset)
                            self._debug['_m_data']['start'] = self._io.pos()
                            self._m_data = (self._io.read_bytes(self.size)).decode(u"ascii")
                            self._debug['_m_data']['end'] = self._io.pos()
                            self._io.seek(_pos)

                        return self._m_data if hasattr(self, '_m_data') else None


                class MetadataAreaDescriptor(KaitaiStruct):
                    SEQ_FIELDS = ["offset", "size"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['offset']['start'] = self._io.pos()
                        self.offset = self._io.read_u8le()
                        self._debug['offset']['end'] = self._io.pos()
                        self._debug['size']['start'] = self._io.pos()
                        self.size = self._io.read_u8le()
                        self._debug['size']['end'] = self._io.pos()

                    @property
                    def data(self):
                        if hasattr(self, '_m_data'):
                            return self._m_data if hasattr(self, '_m_data') else None

                        if self.size != 0:
                            _pos = self._io.pos()
                            self._io.seek(self.offset)
                            self._debug['_m_data']['start'] = self._io.pos()
                            self._raw__m_data = self._io.read_bytes(self.size)
                            _io__raw__m_data = KaitaiStream(BytesIO(self._raw__m_data))
                            self._m_data = Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataArea(_io__raw__m_data, self, self._root)
                            self._m_data._read()
                            self._debug['_m_data']['end'] = self._io.pos()
                            self._io.seek(_pos)

                        return self._m_data if hasattr(self, '_m_data') else None


                class MetadataArea(KaitaiStruct):
                    """According to `[REDHAT]` the metadata area is a circular buffer. New metadata is appended to the old metadata and then the pointer to the start of it is updated. The metadata area, therefore, can contain copies of older versions of the metadata."""
                    SEQ_FIELDS = ["header"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['header']['start'] = self._io.pos()
                        self.header = Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataArea.MetadataAreaHeader(self._io, self, self._root)
                        self.header._read()
                        self._debug['header']['end'] = self._io.pos()

                    class MetadataAreaHeader(KaitaiStruct):
                        SEQ_FIELDS = ["checksum", "signature", "version", "metadata_area_offset", "metadata_area_size", "raw_location_descriptors"]
                        def __init__(self, _io, _parent=None, _root=None):
                            self._io = _io
                            self._parent = _parent
                            self._root = _root if _root else self
                            self._debug = collections.defaultdict(dict)

                        def _read(self):
                            self._debug['checksum']['start'] = self._io.pos()
                            self.checksum = Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataArea.MetadataAreaHeader(self._io, self, self._root)
                            self.checksum._read()
                            self._debug['checksum']['end'] = self._io.pos()
                            self._debug['signature']['start'] = self._io.pos()
                            self.signature = self._io.read_bytes(16)
                            self._debug['signature']['end'] = self._io.pos()
                            if not self.signature == b"\x20\x4C\x56\x4D\x32\x20\x78\x5B\x35\x41\x25\x72\x30\x4E\x2A\x3E":
                                raise kaitaistruct.ValidationNotEqualError(b"\x20\x4C\x56\x4D\x32\x20\x78\x5B\x35\x41\x25\x72\x30\x4E\x2A\x3E", self.signature, self._io, u"/types/physical_volume/types/label/types/volume_header/types/metadata_area/types/metadata_area_header/seq/1")
                            self._debug['version']['start'] = self._io.pos()
                            self.version = self._io.read_u4le()
                            self._debug['version']['end'] = self._io.pos()
                            self._debug['metadata_area_offset']['start'] = self._io.pos()
                            self.metadata_area_offset = self._io.read_u8le()
                            self._debug['metadata_area_offset']['end'] = self._io.pos()
                            self._debug['metadata_area_size']['start'] = self._io.pos()
                            self.metadata_area_size = self._io.read_u8le()
                            self._debug['metadata_area_size']['end'] = self._io.pos()
                            self._debug['raw_location_descriptors']['start'] = self._io.pos()
                            self.raw_location_descriptors = []
                            i = 0
                            while True:
                                if not 'arr' in self._debug['raw_location_descriptors']:
                                    self._debug['raw_location_descriptors']['arr'] = []
                                self._debug['raw_location_descriptors']['arr'].append({'start': self._io.pos()})
                                _t_raw_location_descriptors = Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataArea.MetadataAreaHeader.RawLocationDescriptor(self._io, self, self._root)
                                _t_raw_location_descriptors._read()
                                _ = _t_raw_location_descriptors
                                self.raw_location_descriptors.append(_)
                                self._debug['raw_location_descriptors']['arr'][len(self.raw_location_descriptors) - 1]['end'] = self._io.pos()
                                if  ((_.offset != 0) and (_.size != 0) and (_.checksum != 0)) :
                                    break
                                i += 1
                            self._debug['raw_location_descriptors']['end'] = self._io.pos()

                        class RawLocationDescriptor(KaitaiStruct):
                            """The data area size can be 0. It is assumed it represents the remaining  available data."""

                            class RawLocationDescriptorFlags(Enum):
                                raw_location_ignored = 1
                            SEQ_FIELDS = ["offset", "size", "checksum", "flags"]
                            def __init__(self, _io, _parent=None, _root=None):
                                self._io = _io
                                self._parent = _parent
                                self._root = _root if _root else self
                                self._debug = collections.defaultdict(dict)

                            def _read(self):
                                self._debug['offset']['start'] = self._io.pos()
                                self.offset = self._io.read_u8le()
                                self._debug['offset']['end'] = self._io.pos()
                                self._debug['size']['start'] = self._io.pos()
                                self.size = self._io.read_u8le()
                                self._debug['size']['end'] = self._io.pos()
                                self._debug['checksum']['start'] = self._io.pos()
                                self.checksum = self._io.read_u4le()
                                self._debug['checksum']['end'] = self._io.pos()
                                self._debug['flags']['start'] = self._io.pos()
                                self.flags = KaitaiStream.resolve_enum(Lvm2.PhysicalVolume.Label.VolumeHeader.MetadataArea.MetadataAreaHeader.RawLocationDescriptor.RawLocationDescriptorFlags, self._io.read_u4le())
                                self._debug['flags']['end'] = self._io.pos()


                        @property
                        def metadata(self):
                            if hasattr(self, '_m_metadata'):
                                return self._m_metadata if hasattr(self, '_m_metadata') else None

                            _pos = self._io.pos()
                            self._io.seek(self.metadata_area_offset)
                            self._debug['_m_metadata']['start'] = self._io.pos()
                            self._m_metadata = self._io.read_bytes(self.metadata_area_size)
                            self._debug['_m_metadata']['end'] = self._io.pos()
                            self._io.seek(_pos)
                            return self._m_metadata if hasattr(self, '_m_metadata') else None






    @property
    def sector_size(self):
        if hasattr(self, '_m_sector_size'):
            return self._m_sector_size if hasattr(self, '_m_sector_size') else None

        self._m_sector_size = 512
        return self._m_sector_size if hasattr(self, '_m_sector_size') else None





import base64
from io import BytesIO

from PIL import Image, ImageDraw

from .polyfile import submatcher, InvalidMatch, Match, Submatch


def parse_ines_header(header, parent=None):
    magic = header[:4]
    if magic != b'NES\x1A':
        raise InvalidMatch(f"Expected \"NES\\x1A\" magic at the beginning of the file, but found {magic!r}")
    parent = Submatch(
        name='iNESHeader',
        match_obj=header,
        relative_offset=0,
        length=16,
        parent=parent
    )
    yield parent
    yield Submatch(
        name='iNESMagic',
        match_obj="NES",
        relative_offset=0,
        length=4,
        parent=parent
    )
    prg_size = header[4]
    yield Submatch(
        name='PRGSize',
        match_obj=prg_size,
        relative_offset=4,
        length=1,
        parent=parent
    )
    chr_size = header[5]
    yield Submatch(
        name='CHRSize',
        match_obj=chr_size,
        relative_offset=5,
        length=1,
        parent=parent
    )
    for i in range(6,8):
        yield Submatch(
            name='Flags',
            display_name=f'Flags{i}',
            match_obj=header[i],
            relative_offset=i,
            length=1,
            parent=parent
        )
    yield Submatch(
        name='PRGRAMSize',
        match_obj=header[8],
        relative_offset=8,
        length=1,
        parent=parent
    )
    for i in range(9,11):
        yield Submatch(
            name='Flags',
            display_name=f'Flags{i}',
            match_obj=header[i],
            relative_offset=i,
            length=1,
            parent=parent
        )
    yield Submatch(
        name='UnusedPadding',
        match_obj=header[11:],
        relative_offset=11,
        length=5,
        parent=parent
    )


def parse_ines(file_stream, parent=None):
    header = file_stream.read(16)
    yield from parse_ines_header(header, parent)
    has_trainer = (header[6] & 0b100) != 0
    if has_trainer:
        yield Submatch(
            name='Trainer',
            match_obj=file_stream.read(512),
            relative_offset=16,
            length=512,
            parent=parent
        )
    prg_size = header[4]
    for i in range(prg_size):
        offset = file_stream.tell()
        file_stream.read(16384)
        yield Submatch(
            name='PRGBank',
            display_name=f'PRGBank{i}',
            match_obj='',
            relative_offset=offset,
            length=16384,
            parent=parent
        )
    chr_size = header[5]
    for i in range(chr_size):
        offset = file_stream.tell()
        chr_bytes = file_stream.read(8192)
        chr_img = render_chr(chr_bytes)
        with BytesIO() as img_data:
            chr_img.save(img_data, "PNG")
            yield Submatch(
                name='CHRBank',
                display_name=f'CHRBank{i}',
                match_obj='',
                relative_offset=offset,
                length=8192,
                parent=parent,
                img_data=f"data:image/png;base64,{base64.b64encode(img_data.getvalue()).decode('utf-8')}"
            )


def chr_values(chr_bytes: bytes):
    for i, offset in enumerate(range(0, len(chr_bytes), 16)):
        base_x = (i % 16) * 8
        base_y = (i // 16) * 8
        for y in range(8):
            for x in range(8):
                shift = 7 - x
                yield base_x + x,\
                      base_y + y,\
                      ((((chr_bytes[offset + y + 8] >> shift) & 0b1)) << 1) | ((chr_bytes[offset + y] >> shift) & 0b1)


def render_chr(chr_bytes: bytes) -> Image:
    img = Image.new(mode='L', size=(8*16, 8*32))
    d = ImageDraw.Draw(img)
    for x, y, pixel in chr_values(chr_bytes):
        if pixel == 1:
            pixel = 0xFF//3
        elif pixel == 2:
            pixel = 0xFF//3*2
        elif pixel == 3:
            pixel = 0xFF
        d.point((x, y), pixel)
    return img


@submatcher("application/x-nes-rom")
class INESMatcher(Match):
    def submatch(self, file_stream):
        yield from parse_ines(file_stream, parent=self)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AppleSingleDouble(KaitaiStruct):
    """AppleSingle and AppleDouble files are used by certain Mac
    applications (e.g. Finder) to store Mac-specific file attributes on
    filesystems that do not support that.
    
    Syntactically, both formats are the same, the only difference is how
    they are being used:
    
    * AppleSingle means that only one file will be created on external
      filesystem that will hold both the data (AKA "data fork" in Apple
      terminology), and the attributes (AKA "resource fork").
    * AppleDouble means that two files will be created: a normal file
      that keeps the data ("data fork") is kept separately from an
      auxiliary file that contains attributes ("resource fork"), which
      is kept with the same name, but starting with an extra dot and
      underscore `._` to keep it hidden.
    
    In modern practice (Mac OS X), Finder only uses AppleDouble to keep
    compatibility with other OSes, as virtually nobody outside of Mac
    understands how to access data in AppleSingle container.
    
    .. seealso::
       Source - http://kaiser-edv.de/documents/AppleSingle_AppleDouble.pdf
    """

    class FileType(Enum):
        apple_single = 333312
        apple_double = 333319
    SEQ_FIELDS = ["magic", "version", "reserved", "num_entries", "entries"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = KaitaiStream.resolve_enum(AppleSingleDouble.FileType, self._io.read_u4be())
        self._debug['magic']['end'] = self._io.pos()
        self._debug['version']['start'] = self._io.pos()
        self.version = self._io.read_u4be()
        self._debug['version']['end'] = self._io.pos()
        self._debug['reserved']['start'] = self._io.pos()
        self.reserved = self._io.read_bytes(16)
        self._debug['reserved']['end'] = self._io.pos()
        self._debug['num_entries']['start'] = self._io.pos()
        self.num_entries = self._io.read_u2be()
        self._debug['num_entries']['end'] = self._io.pos()
        self._debug['entries']['start'] = self._io.pos()
        self.entries = [None] * (self.num_entries)
        for i in range(self.num_entries):
            if not 'arr' in self._debug['entries']:
                self._debug['entries']['arr'] = []
            self._debug['entries']['arr'].append({'start': self._io.pos()})
            _t_entries = AppleSingleDouble.Entry(self._io, self, self._root)
            _t_entries._read()
            self.entries[i] = _t_entries
            self._debug['entries']['arr'][i]['end'] = self._io.pos()

        self._debug['entries']['end'] = self._io.pos()

    class Entry(KaitaiStruct):

        class Types(Enum):
            data_fork = 1
            resource_fork = 2
            real_name = 3
            comment = 4
            icon_bw = 5
            icon_color = 6
            file_dates_info = 8
            finder_info = 9
            macintosh_file_info = 10
            prodos_file_info = 11
            msdos_file_info = 12
            afp_short_name = 13
            afp_file_info = 14
            afp_directory_id = 15
        SEQ_FIELDS = ["type", "ofs_body", "len_body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(AppleSingleDouble.Entry.Types, self._io.read_u4be())
            self._debug['type']['end'] = self._io.pos()
            self._debug['ofs_body']['start'] = self._io.pos()
            self.ofs_body = self._io.read_u4be()
            self._debug['ofs_body']['end'] = self._io.pos()
            self._debug['len_body']['start'] = self._io.pos()
            self.len_body = self._io.read_u4be()
            self._debug['len_body']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek(self.ofs_body)
            self._debug['_m_body']['start'] = self._io.pos()
            _on = self.type
            if _on == AppleSingleDouble.Entry.Types.finder_info:
                self._raw__m_body = self._io.read_bytes(self.len_body)
                _io__raw__m_body = KaitaiStream(BytesIO(self._raw__m_body))
                self._m_body = AppleSingleDouble.FinderInfo(_io__raw__m_body, self, self._root)
                self._m_body._read()
            else:
                self._m_body = self._io.read_bytes(self.len_body)
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class FinderInfo(KaitaiStruct):
        """Information specific to Finder.
        
        .. seealso::
           older Inside Macintosh, Volume II page 84 or Volume IV page 104.
        """
        SEQ_FIELDS = ["file_type", "file_creator", "flags", "location", "folder_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['file_type']['start'] = self._io.pos()
            self.file_type = self._io.read_bytes(4)
            self._debug['file_type']['end'] = self._io.pos()
            self._debug['file_creator']['start'] = self._io.pos()
            self.file_creator = self._io.read_bytes(4)
            self._debug['file_creator']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u2be()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['location']['start'] = self._io.pos()
            self.location = AppleSingleDouble.Point(self._io, self, self._root)
            self.location._read()
            self._debug['location']['end'] = self._io.pos()
            self._debug['folder_id']['start'] = self._io.pos()
            self.folder_id = self._io.read_u2be()
            self._debug['folder_id']['end'] = self._io.pos()


    class Point(KaitaiStruct):
        """Specifies 2D coordinate in QuickDraw grid."""
        SEQ_FIELDS = ["x", "y"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_u2be()
            self._debug['x']['end'] = self._io.pos()
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_u2be()
            self._debug['y']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Pcx(KaitaiStruct):
    """PCX is a bitmap image format originally used by PC Paintbrush from
    ZSoft Corporation. Originally, it was a relatively simple 128-byte
    header + uncompressed bitmap format, but latest versions introduced
    more complicated palette support and RLE compression.
    
    There's an option to encode 32-bit or 16-bit RGBA pixels, and thus
    it can potentially include transparency. Theoretically, it's
    possible to encode resolution or pixel density in the some of the
    header fields too, but in reality there's no uniform standard for
    these, so different implementations treat these differently.
    
    PCX format was never made a formal standard. "ZSoft Corporation
    Technical Reference Manual" for "Image File (.PCX) Format", last
    updated in 1991, is likely the closest authoritative source.
    
    .. seealso::
       Source - https://web.archive.org/web/20100206055706/http://www.qzx.com/pc-gpe/pcx.txt
    """

    class Versions(Enum):
        v2_5 = 0
        v2_8_with_palette = 2
        v2_8_without_palette = 3
        paintbrush_for_windows = 4
        v3_0 = 5

    class Encodings(Enum):
        rle = 1
    SEQ_FIELDS = ["hdr"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['hdr']['start'] = self._io.pos()
        self._raw_hdr = self._io.read_bytes(128)
        _io__raw_hdr = KaitaiStream(BytesIO(self._raw_hdr))
        self.hdr = Pcx.Header(_io__raw_hdr, self, self._root)
        self.hdr._read()
        self._debug['hdr']['end'] = self._io.pos()

    class Header(KaitaiStruct):
        """
        .. seealso::
           - "ZSoft .PCX FILE HEADER FORMAT" - https://web.archive.org/web/20100206055706/http://www.qzx.com/pc-gpe/pcx.txt
        """
        SEQ_FIELDS = ["magic", "version", "encoding", "bits_per_pixel", "img_x_min", "img_y_min", "img_x_max", "img_y_max", "hdpi", "vdpi", "palette_16", "reserved", "num_planes", "bytes_per_line", "palette_info", "h_screen_size", "v_screen_size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(1)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x0A":
                raise kaitaistruct.ValidationNotEqualError(b"\x0A", self.magic, self._io, u"/types/header/seq/0")
            self._debug['version']['start'] = self._io.pos()
            self.version = KaitaiStream.resolve_enum(Pcx.Versions, self._io.read_u1())
            self._debug['version']['end'] = self._io.pos()
            self._debug['encoding']['start'] = self._io.pos()
            self.encoding = KaitaiStream.resolve_enum(Pcx.Encodings, self._io.read_u1())
            self._debug['encoding']['end'] = self._io.pos()
            self._debug['bits_per_pixel']['start'] = self._io.pos()
            self.bits_per_pixel = self._io.read_u1()
            self._debug['bits_per_pixel']['end'] = self._io.pos()
            self._debug['img_x_min']['start'] = self._io.pos()
            self.img_x_min = self._io.read_u2le()
            self._debug['img_x_min']['end'] = self._io.pos()
            self._debug['img_y_min']['start'] = self._io.pos()
            self.img_y_min = self._io.read_u2le()
            self._debug['img_y_min']['end'] = self._io.pos()
            self._debug['img_x_max']['start'] = self._io.pos()
            self.img_x_max = self._io.read_u2le()
            self._debug['img_x_max']['end'] = self._io.pos()
            self._debug['img_y_max']['start'] = self._io.pos()
            self.img_y_max = self._io.read_u2le()
            self._debug['img_y_max']['end'] = self._io.pos()
            self._debug['hdpi']['start'] = self._io.pos()
            self.hdpi = self._io.read_u2le()
            self._debug['hdpi']['end'] = self._io.pos()
            self._debug['vdpi']['start'] = self._io.pos()
            self.vdpi = self._io.read_u2le()
            self._debug['vdpi']['end'] = self._io.pos()
            self._debug['palette_16']['start'] = self._io.pos()
            self.palette_16 = self._io.read_bytes(48)
            self._debug['palette_16']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(1)
            self._debug['reserved']['end'] = self._io.pos()
            if not self.reserved == b"\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00", self.reserved, self._io, u"/types/header/seq/11")
            self._debug['num_planes']['start'] = self._io.pos()
            self.num_planes = self._io.read_u1()
            self._debug['num_planes']['end'] = self._io.pos()
            self._debug['bytes_per_line']['start'] = self._io.pos()
            self.bytes_per_line = self._io.read_u2le()
            self._debug['bytes_per_line']['end'] = self._io.pos()
            self._debug['palette_info']['start'] = self._io.pos()
            self.palette_info = self._io.read_u2le()
            self._debug['palette_info']['end'] = self._io.pos()
            self._debug['h_screen_size']['start'] = self._io.pos()
            self.h_screen_size = self._io.read_u2le()
            self._debug['h_screen_size']['end'] = self._io.pos()
            self._debug['v_screen_size']['start'] = self._io.pos()
            self.v_screen_size = self._io.read_u2le()
            self._debug['v_screen_size']['end'] = self._io.pos()


    class TPalette256(KaitaiStruct):
        SEQ_FIELDS = ["magic", "colors"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(1)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x0C":
                raise kaitaistruct.ValidationNotEqualError(b"\x0C", self.magic, self._io, u"/types/t_palette_256/seq/0")
            self._debug['colors']['start'] = self._io.pos()
            self.colors = [None] * (256)
            for i in range(256):
                if not 'arr' in self._debug['colors']:
                    self._debug['colors']['arr'] = []
                self._debug['colors']['arr'].append({'start': self._io.pos()})
                _t_colors = Pcx.Rgb(self._io, self, self._root)
                _t_colors._read()
                self.colors[i] = _t_colors
                self._debug['colors']['arr'][i]['end'] = self._io.pos()

            self._debug['colors']['end'] = self._io.pos()


    class Rgb(KaitaiStruct):
        SEQ_FIELDS = ["r", "g", "b"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['r']['start'] = self._io.pos()
            self.r = self._io.read_u1()
            self._debug['r']['end'] = self._io.pos()
            self._debug['g']['start'] = self._io.pos()
            self.g = self._io.read_u1()
            self._debug['g']['end'] = self._io.pos()
            self._debug['b']['start'] = self._io.pos()
            self.b = self._io.read_u1()
            self._debug['b']['end'] = self._io.pos()


    @property
    def palette_256(self):
        """
        .. seealso::
           - "VGA 256 Color Palette Information" - https://web.archive.org/web/20100206055706/http://www.qzx.com/pc-gpe/pcx.txt
        """
        if hasattr(self, '_m_palette_256'):
            return self._m_palette_256 if hasattr(self, '_m_palette_256') else None

        if  ((self.hdr.version == Pcx.Versions.v3_0) and (self.hdr.bits_per_pixel == 8) and (self.hdr.num_planes == 1)) :
            _pos = self._io.pos()
            self._io.seek((self._io.size() - 769))
            self._debug['_m_palette_256']['start'] = self._io.pos()
            self._m_palette_256 = Pcx.TPalette256(self._io, self, self._root)
            self._m_palette_256._read()
            self._debug['_m_palette_256']['end'] = self._io.pos()
            self._io.seek(_pos)

        return self._m_palette_256 if hasattr(self, '_m_palette_256') else None


